from deepfriedmarshmallow.serializer import JitDeserialize, JitSerialize


class JitSchemaMixin:
    jit_serialize_class = JitSerialize
    jit_deserialize_class = JitDeserialize
    _is_jit = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._serialize = self.jit_serialize_class(self)
        self._deserialize = self.jit_deserialize_class(self)



__version__ = "1.0.0beta1"

from deepfriedmarshmallow.import_patch import deep_fry_marshmallow
from deepfriedmarshmallow.jit import JitContext, generate_method_bodies
from deepfriedmarshmallow.mixin import JitSchemaMixin
from deepfriedmarshmallow.patch import deep_fry_schema, deep_fry_schema_object
from deepfriedmarshmallow.serializer import JitDeserialize, JitSerialize


def __getattr__(name):
    if name == "JitSchema":
        if "JitSchema" not in globals() or not globals()["JitSchema"]:
            from marshmallow import Schema

            from deepfriedmarshmallow.mixin import JitSchemaMixin

            class _JitSchemaImpl(JitSchemaMixin, Schema):
                pass

            # Cache and return the created class
            globals()["JitSchema"] = _JitSchemaImpl
        return globals()["JitSchema"]

    msg = f"module '{__name__}' has no attribute {name}"
    raise AttributeError(msg)



import os
from contextlib import contextmanager


class IndentedString:
    """Utility class for printing indented strings via a context manager."""

    def __init__(self, content="", indent=4):
        # type: (Union[str, IndentedString], int) -> None
        self.result = []  # type: List[str]
        self._indent = indent
        self.__indents = [""]
        if content:
            self.__iadd__(content)

    @contextmanager
    def indent(self):
        self.__indents.append(self.__indents[-1] + (self._indent * " "))
        try:
            yield
        finally:
            self.__indents.pop()

    def __iadd__(self, other):
        # type: (Union[str, IndentedString]) -> IndentedString
        if isinstance(other, IndentedString):
            for line in other.result:
                self.result.append(self.__indents[-1] + line)
        else:
            self.result.append(self.__indents[-1] + other)
        return self

    def __str__(self):
        # type: () -> str
        return os.linesep.join(self.result)



from typing import Type

from deepfriedmarshmallow.log import logger
from deepfriedmarshmallow.serializer import JitDeserialize, JitSerialize


def deep_fry_schema_object(schema: "marshmallow.Schema") -> None:
    """Patches a Marshmallow schema object to support JIT compilation."""
    logger.info(
        f"Deep-frying schema {schema.__name__} instance. Current schema is_jit status:"
        f" {getattr(schema, '_is_jit', 'N/A')}"
    )
    schema._is_jit = True

    schema._serialize = JitSerialize(schema)
    schema._deserialize = JitDeserialize(schema)
    schema.__doc__ = "Marshmallow module enhanced with Deep-Fried Marshmallow (via patch)"


def deep_fry_schema(cls: Type["marshmallow.Schema"]) -> None:
    """Patches a Marshmallow schema to support JIT compilation."""
    logger.info(f"Deep-frying schema {cls.__name__}. Current schema is_jit status: {getattr(cls, '_is_jit', 'N/A')}")

    cls._is_jit = True

    super_init = cls.__init__

    def new_init(self, *args, **kwargs):
        super_init(self, *args, **kwargs)
        self._serialize = JitSerialize(self)
        self._deserialize = JitDeserialize(self)

    cls.__init__ = new_init
    cls.__doc__ = "Marshmallow module enhanced with Deep-Fried Marshmallow (via patch)"



import logging
import os


logger = logging.getLogger("DeepFriedMarshmallow")
sh = logging.StreamHandler()
sh.setFormatter(logging.Formatter("[%(asctime)s] [%(levelname)s] [Deep-Fried Marshmallow] %(message)s"))

if logger.level == logging.NOTSET:
    try:
        logger.setLevel(os.getenv("DFM_LOG_LEVEL", logging.WARN))
        sh.setLevel(os.getenv("DFM_LOG_LEVEL", logging.WARN))
    except ValueError:
        logger.setLevel(logging.WARN)
        sh.setLevel(logging.WARN)

logger.addHandler(sh)



from deepfriedmarshmallow.jit import (
    JitContext,
    generate_deserialize_method,
    generate_serialize_method,
)
from deepfriedmarshmallow.log import logger


class JitMethodWrapper:
    def __init__(self, schema, method):
        self._schema = schema
        self._method = method

        self._jit_method = None
        self._prev_fields_dict = None

    def __call__(self, obj, many=False, **kwargs):  # noqa: FBT002
        self._ensure_jit_method()

        logger.debug(f"JIT method called with {obj=}")
        try:
            result = self._jit_method(obj, many=many)
            logger.debug(f"JIT method succeeded for {obj=} with {result=}")
        except Exception as e:
            logger.warning(f"JIT method failed, falling back to non-JIT method: {e}", exc_info=e)
            result = self._method(obj, many=many, **kwargs)
            logger.debug(f"Fallback method succeeded for {obj=} with {result=}")

        return result

    def _ensure_jit_method(self):
        if self._jit_method is None:
            logger.debug(f"Generating JIT method {self._method} for {self._schema}")
            self._jit_method = self.generate_jit_method(self._schema, JitContext())

    def generate_jit_method(self, schema, context):
        raise NotImplementedError

    def __getattr__(self, item):
        return getattr(self._method, item)


class JitSerialize(JitMethodWrapper):
    def __init__(self, schema):
        super().__init__(schema, schema._serialize)

    def generate_jit_method(self, schema, context):
        return generate_serialize_method(schema, context)


class JitDeserialize(JitMethodWrapper):
    def __init__(self, schema):
        super().__init__(schema, schema._deserialize)

    def generate_jit_method(self, schema, context):
        return generate_deserialize_method(schema, context)



def is_overridden(instance_func, class_func):
    # type: (MethodType, MethodType) -> bool
    return instance_func.__func__ is not class_func


def is_schema_overridden(schema: "marshmallow.Schema") -> bool:
    return hasattr(schema, "_is_jit") and schema._is_jit



import importlib
import sys
import types

from deepfriedmarshmallow.log import logger


class MarshmallowImportReplacer:
    def __init__(self):
        logger.debug("Marshmallow import hook created")
        # A flag to prevent recursive importing
        self.is_importing = False

    def find_spec(self, name, path, target=None):
        if name == "marshmallow" and not self.is_importing:
            return importlib.util.spec_from_loader(name, loader=self, origin=path)  # Use self here
        return None

    def create_module(self, spec):
        logger.debug("Marshmallow module created")
        # Set flag to avoid recursive importing
        self.is_importing = True

        # Import original marshmallow and create a new module
        original_marshmallow = importlib.import_module("marshmallow")

        # Manually create a new module instead of using importlib.util.module_from_spec
        new_module = types.ModuleType(spec.name)
        new_module.__dict__.update(original_marshmallow.__dict__)

        # Release flag as import is done
        self.is_importing = False

        # Patch marshmallow module
        return self.patch_marshmallow_module(new_module)

    def exec_module(self, module):
        # No action needed here, as all actions are performed in create_module
        pass

    @staticmethod
    def patch_marshmallow_module(module):
        if hasattr(module, "Schema") and hasattr(module.Schema, "_is_jit") and module.Schema._is_jit:
            # Marshmallow has already been patched, return the module as-is
            return module

        from deepfriedmarshmallow import JitSchema

        module.Schema = JitSchema
        module.__doc__ = "Marshmallow module enhanced with Deep-Fried Marshmallow (Schema class replacement)"
        return module


def deep_fry_marshmallow():
    if "marshmallow" in sys.modules:
        # Marshmallow has already been imported, update the existing module
        logger.info("Marshmallow has already been imported, updating the existing module")
        sys.modules["marshmallow"] = MarshmallowImportReplacer.patch_marshmallow_module(sys.modules["marshmallow"])
    else:
        # Marshmallow has not been imported yet, insert the import hook
        logger.info("Marshmallow has not been imported yet, inserting the import hook")
        sys.meta_path.insert(0, MarshmallowImportReplacer())
